# Author Yechen Qiao
# Common Molecule Utilities for Molecule Transfers with Sapio
from typing import cast

from rdkit import Chem
from rdkit.Chem import Crippen, MolToInchi
from rdkit.Chem import Descriptors
from rdkit.Chem import rdMolDescriptors
from rdkit.Chem.EnumerateStereoisomers import StereoEnumerationOptions, EnumerateStereoisomers
from rdkit.Chem.MolStandardize import rdMolStandardize
from rdkit.Chem.SaltRemover import SaltRemover
from rdkit.Chem.rdChemReactions import ChemicalReaction
from rdkit.Chem.rdchem import Mol, RWMol, Bond

from sapiopycommons.chem.IndigoMolecules import indigo, renderer, indigo_inchi

metal_disconnector = rdMolStandardize.MetalDisconnector()
tautomer_params = Chem.MolStandardize.rdMolStandardize.CleanupParameters()
tautomer_params.tautomerRemoveSp3Stereo = False
tautomer_params.tautomerRemoveBondStereo = False
tautomer_params.tautomerReassignStereo = False
tautomer_params.tautomerRemoveIsotopicHs = True
enumerator = rdMolStandardize.TautomerEnumerator(tautomer_params)


def remove_dative_bonds_from_mol(mol: Mol) -> RWMol:
    """
    Create a new copy of RWMol molecule and remove all dative bonds in the molecule.
    :param mol: The original molecule
    :return: The new molecule with dative bonds removed.
    """
    ret: RWMol = Chem.RWMol(mol)
    bonds_to_remove = []
    bond: Bond
    for bond in ret.GetBonds():
        if bond.GetBondType() in [Chem.BondType.DATIVER, Chem.BondType.DATIVE, Chem.BondType.DATIVEL,
                                  Chem.BondType.DATIVEONE]:
            bonds_to_remove.append((bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()))
    for atom1_idx, atom2_idx in bonds_to_remove:
        ret.RemoveBond(atom1_idx, atom2_idx)
    return ret


def get_enhanced_stereo_reg_hash(mol: Mol, enhanced_stereo: bool) -> str:
    """
    Get the Registration Hash for the molecule by the current registration configuration.
    When we are running if we are canonicalization of tautomers or cleaning up any other way, do they first before calling.
    :param mol: The molecule to obtain hash for.
    :param canonical_tautomer: Whether the registry system canonicalize the tautomers.
    :param enhanced_stereo: Whether we are computing enhanced stereo at all.
    :return: The enhanced stereo hash.
    """
    if enhanced_stereo:
        from rdkit.Chem.RegistrationHash import GetMolLayers, GetMolHash, HashScheme
        layers = GetMolLayers(mol, enable_tautomer_hash_v2=True)
        hash_scheme: HashScheme = HashScheme.TAUTOMER_INSENSITIVE_LAYERS
        return GetMolHash(layers, hash_scheme=hash_scheme)
    else:
        return ""


def neutralize_atoms(mol) -> Mol:
    """
    Neutralize atoms per https://baoilleach.blogspot.com/2019/12/no-charge-simple-approach-to.html
    """
    pattern = Chem.MolFromSmarts("[+1!h0!$([*]~[-1,-2,-3,-4]),-1!$([*]~[+1,+2,+3,+4])]")
    at_matches = mol.GetSubstructMatches(pattern)
    at_matches_list = [y[0] for y in at_matches]
    if len(at_matches_list) > 0:
        for at_idx in at_matches_list:
            atom = mol.GetAtomWithIdx(at_idx)
            chg = atom.GetFormalCharge()
            hcount = atom.GetTotalNumHs()
            atom.SetFormalCharge(0)
            atom.SetNumExplicitHs(hcount - chg)
            atom.UpdatePropertyCache()
    return mol


def find_all_possible_stereoisomers(m: Mol, only_unassigned=True, try_embedding=True, unique=True, max_isomers=200) \
        -> list[Mol]:
    """
    Find all possible candidates of stereoisomers given the current molecule.

    :param m: The molecule to search for.
    :param only_unassigned: Whether to only permute on unspecified stereocenter.
    :param try_embedding: if set the process attempts to generate a standard RDKit distance geometry conformation for
    the stereisomer.
    If this fails, we assume that the stereoisomer is non-physical and don't return it.
    :param unique: whether to remove duplicates by isomer identity.
    :param max_isomers: Maximum number of search results to return.
    """
    # noinspection PyBroadException
    try:
        opts = StereoEnumerationOptions(tryEmbedding=try_embedding, unique=unique, onlyUnassigned=only_unassigned,
                                        maxIsomers=max_isomers)
        return list(EnumerateStereoisomers(m, options=opts))
    except:
        return []


def has_chiral_centers(m: Mol):
    """
    Returns true iff the molecule provided has at least 1 chiral centers (when stereochemistry is relevant).

    :param m: The molecule to test.
    """
    # noinspection PyBroadException
    try:
        chiral_centers: list = Chem.FindMolChiralCenters(m, force=True, includeUnassigned=True,
                                                         useLegacyImplementation=False)
        return len(chiral_centers) > 0
    except:
        return False


def mol_to_img(mol_str: str) -> str:
    """
    Convert molecule into image.

    :param mol_str: The molecule INCHI.
    :return: The SVG image text.
    """
    mol = indigo.loadMolecule(mol_str)
    return renderer.renderToString(mol)


def mol_to_sapio_partial_pojo(mol: Mol):
    """
    Get the minimum information about molecule to Sapio, just its SMILES, V3000, and image data.

    :param mol: The molecule to read the simplified data from.
    """
    Chem.SanitizeMol(mol)
    mol.UpdatePropertyCache()
    smiles = Chem.MolToSmiles(mol)
    molBlock = Chem.MolToMolBlock(mol, forceV3000=True)
    img = mol_to_img(mol)
    molecule = dict()
    molecule["smiles"] = smiles
    molecule["molBlock"] = molBlock
    molecule["image"] = img
    return molecule


def get_cxs_smiles_hash(mol: Mol, enhanced_stereo: bool) -> str:
    """
    Return the SHA1 CXS Smiles hash for the canonical, isomeric CXS SMILES of the molecule.
    """
    if not enhanced_stereo:
        return ""
    import hashlib
    return hashlib.sha1(Chem.MolToCXSmiles(mol, canonical=True, isomericSmiles=True).encode()).hexdigest()


def get_has_or_group(mol: Mol, enhanced_stereo: bool) -> bool:
    """
    Return true if and only if: enhanced stereochemistry is enabled and there is at least one OR group in mol.
    """
    if not enhanced_stereo:
        return False
    from rdkit.Chem import StereoGroup_vect, STEREO_OR
    stereo_groups: StereoGroup_vect = mol.GetStereoGroups()
    for stereo_group in stereo_groups:
        if stereo_group.GetGroupType() == STEREO_OR:
            return True
    return False


def mol_to_sapio_substance(mol: Mol, include_stereoisomers=False,
                           normalize: bool = False, remove_salt: bool = False, make_images: bool = False,
                           salt_def: str | None = None, canonical_tautomer: bool = True,
                           enhanced_stereo: bool = False, remove_atom_map: bool = True):
    """
    Convert a molecule in RDKit to a molecule POJO in Sapio.

    :param mol: The molecule in RDKit.
    :param normalize If true, will normalize the functional groups and return normalized result.
    :param remove_salt If true, we will remove salts iteratively from the molecule before returning their data.
    We will also populate desaltedList with molecules we deleted.
    :param make_images Whether to make images as part of the result without having another script to resolve it.
    :param salt_def: if not none, specifies custom salt to be used during the desalt process.
    :param canonical_tautomer: if True, we will attempt to compute canonical tautomer for the molecule. Slow!
    This is needed for a registry. Note it stops after enumeration of 1000.
    :param enhanced_stereo: If enabled, enhanced stereo hash will be produced.
    :param remove_atom_map: When set, clear all atom AAM maps that were set had it been merged into some reactions earlier.
    :return: The molecule POJO for Sapio.
    """
    molecule = dict()
    if remove_atom_map:
        [a.SetAtomMapNum(0) for a in mol.GetAtoms()]
    Chem.SanitizeMol(mol)
    mol.UpdatePropertyCache()
    Chem.GetSymmSSSR(mol)
    if normalize:
        try:
            mol = Chem.RemoveHs(mol)
            mol = metal_disconnector.Disconnect(mol)
            mol = rdMolStandardize.Normalize(mol)
            molecule["normError"] = ""
        except Exception as e:
            molecule["normError"] = str(e)
    if remove_salt:
        try:
            remover = SaltRemover(defnData=salt_def)
            mol, deleted = remover.StripMolWithDeleted(mol)
            molecule["desaltedList"] = [Chem.MolToSmarts(x) for x in deleted]
            molecule["desaltError"] = ""
        except Exception as e:
            molecule["desaltError"] = str(e)
            molecule["desaltedList"] = []
    if normalize or remove_salt:
        mol = neutralize_atoms(mol)
    #//CR-46021 Jarvis: no canonicalize tautomers.
    if canonical_tautomer:
        mol = enumerator.Canonicalize(mol)
    Chem.SanitizeMol(mol)
    mol.UpdatePropertyCache()
    Chem.GetSymmSSSR(mol)
    smiles = Chem.MolToSmiles(mol)
    cLogP = Crippen.MolLogP(mol)
    tpsa = Descriptors.TPSA(mol)
    amw = Descriptors.MolWt(mol)
    exactMass = Descriptors.ExactMolWt(mol)
    molFormula = rdMolDescriptors.CalcMolFormula(mol)
    charge = Chem.GetFormalCharge(mol)
    molBlock = Chem.MolToMolBlock(mol, forceV3000=True)

    molecule["cLogP"] = cLogP
    molecule["tpsa"] = tpsa
    molecule["amw"] = amw
    molecule["exactMass"] = exactMass
    molecule["molFormula"] = molFormula
    molecule["charge"] = charge
    molecule["numHBondAcceptors"] = rdMolDescriptors.CalcNumHBA(mol)
    # This is number of H-Bond Donor
    molecule["numHBonds"] = rdMolDescriptors.CalcNumHBD(mol)
    molecule["molBlock"] = molBlock
    # Create a copy of molecule before modifying it for InChI generation.
    inchi_mol: Mol = remove_dative_bonds_from_mol(mol)
    rdkit_inchi = MolToInchi(inchi_mol)
    # If INCHI is completely invalid, we fail this molecule.
    if not rdkit_inchi:
        MolToInchi(mol, treatWarningAsError=True)
    if make_images:
        img = mol_to_img(smiles)
        molecule["image"] = img
    else:
        molecule["image"] = None
    # We need to test the INCHI can be loaded back to indigo.
    indigo_mol = indigo.loadMolecule(molBlock)
    indigo_mol.aromatize()
    if enhanced_stereo:
        # Remove enhanced stereo layer when generating InChI as the stereo hash is generated separately for reg.
        Chem.CanonicalizeEnhancedStereo(inchi_mol)
        molecule["inchi"] = Chem.MolToInchi(inchi_mol)
        molecule["inchiKey"] = Chem.MolToInchiKey(inchi_mol)
    else:
        indigo_inchi.resetOptions()
        indigo_inchi_mol = indigo.loadMolecule(Chem.MolToMolBlock(inchi_mol, forceV3000=True))
        indigo_inchi_str = indigo_inchi.getInchi(indigo_inchi_mol)
        molecule["inchi"] = indigo_inchi_str
        indigo_inchi_key_str = indigo_inchi.getInchiKey(indigo_inchi_str)
        molecule["inchiKey"] = indigo_inchi_key_str
    molecule["smiles"] = indigo_mol.smiles()
    molecule["reg_hash"] = get_enhanced_stereo_reg_hash(mol, enhanced_stereo=enhanced_stereo)
    molecule["cxsmiles_hash"] = get_cxs_smiles_hash(mol, enhanced_stereo=enhanced_stereo)
    molecule["has_or_group"] = get_has_or_group(mol, enhanced_stereo=enhanced_stereo)
    return molecule


def mol_to_sapio_compound(mol: Mol, include_stereoisomers=False, enhanced_stereo: bool = False,
                          salt_def: str | None = None, resolve_canonical: bool = True,
                          make_images: bool = False, canonical_tautomer: bool = True,
                          remove_atom_map: bool = True):
    ret = dict()
    ret['originalMol'] = mol_to_sapio_substance(mol, include_stereoisomers=False,
                                                normalize=False, remove_salt=False, make_images=make_images,
                                                canonical_tautomer=canonical_tautomer,
                                                enhanced_stereo=enhanced_stereo, remove_atom_map=remove_atom_map)
    if resolve_canonical:
        ret['canonicalMol'] = mol_to_sapio_substance(mol, include_stereoisomers=False,
                                                     normalize=True, remove_salt=True, make_images=make_images,
                                                     salt_def=salt_def, canonical_tautomer=canonical_tautomer,
                                                     enhanced_stereo=enhanced_stereo, remove_atom_map=remove_atom_map)
    return ret



from typing import Iterable

from sapiopylib.rest.pojo.CustomReport import RawTermOperation, CompositeTermOperation, RawReportTerm, \
    CompositeReportTerm, AbstractReportTerm, FieldCompareReportTerm

from sapiopycommons.general.aliases import DataTypeIdentifier, AliasUtil, FieldIdentifier

# Raw term operations, for comparing field values.
EQ = RawTermOperation.EQUAL_TO_OPERATOR
NEQ = RawTermOperation.NOT_EQUAL_TO_OPERATOR
LT = RawTermOperation.LESS_THAN_OPERATOR
LTE = RawTermOperation.LESS_THAN_OR_EQUAL_OPERATOR
GT = RawTermOperation.GREATER_THAN_OPERATOR
GTE = RawTermOperation.GREATER_THAN_OR_EQUAL_OPERATOR

# Composite term operations, for comparing two terms.
AND = CompositeTermOperation.AND_OPERATOR
OR = CompositeTermOperation.OR_OPERATOR

# Forms that field term values can take.
TermValue = str | int | float | bool | Iterable | None


class TermBuilder:
    """
    A class that allows for the easier constructions of custom report terms.
    """
    data_type: str

    def __init__(self, data_type: DataTypeIdentifier):
        """
        :param data_type: An object that can be used to identify a data type name. Used as the default data type name
            of the terms created from this TermBuilder.
        """
        self.data_type = AliasUtil.to_data_type_name(data_type)

    def all_records_term(self, *, data_type: DataTypeIdentifier | None = None) -> RawReportTerm:
        """
        Create a raw report term that captures all records of the given data type.

        :param data_type: If provided, the data type of this term. If not provided, then the data type that this term
            builder was instantiated with is used.
        :return: A raw report term for "data_type.RecordId >= 0".
        """
        data_type: str = AliasUtil.to_data_type_name(data_type) if data_type else self.data_type
        return RawReportTerm(data_type, "RecordId", GTE, TermBuilder.to_term_val(0))

    def is_term(self, field: FieldIdentifier, value: TermValue,
                *, data_type: DataTypeIdentifier | None = None, trim: bool = False) -> RawReportTerm:
        """
        Create a raw report term for comparing a field value with an equals operation.

        :param field: The data field of this term.
        :param value: The value to compare for this term.
        :param data_type: If provided, the data type of this term. If not provided, then the data type that this term
            builder was instantiated with is used.
        :param trim: Whether the string of the given value should be trimmed of trailing and leading whitespace.
        :return: A raw report term for "data_type.field = value".
        """
        data_type: str = AliasUtil.to_data_type_name(data_type) if data_type else self.data_type
        return RawReportTerm(AliasUtil.to_data_type_name(data_type), AliasUtil.to_data_field_name(field), EQ,
                             TermBuilder.to_term_val(value), trim)

    def not_term(self, field: FieldIdentifier, value: TermValue,
                 *, data_type: DataTypeIdentifier | None = None, trim: bool = False) -> RawReportTerm:
        """
        Create a raw report term for comparing a field value with a not equals operation.

        :param field: The data field of this term.
        :param value: The value to compare for this term.
        :param data_type: If provided, the data type of this term. If not provided, then the data type that this term
            builder was instantiated with is used.
        :param trim: Whether the string of the given value should be trimmed of trailing and leading whitespace.
        :return: A raw report term for "data_type.field != value".
        """
        data_type: str = AliasUtil.to_data_type_name(data_type) if data_type else self.data_type
        return RawReportTerm(AliasUtil.to_data_type_name(data_type), AliasUtil.to_data_field_name(field), NEQ,
                             TermBuilder.to_term_val(value), trim)

    def lt_term(self, field: FieldIdentifier, value: TermValue,
                *, data_type: DataTypeIdentifier | None = None, trim: bool = False) -> RawReportTerm:
        """
        Create a raw report term for comparing a field value with a less than operation.

        :param field: The data field of this term.
        :param value: The value to compare for this term.
        :param data_type: If provided, the data type of this term. If not provided, then the data type that this term
            builder was instantiated with is used.
        :param trim: Whether the string of the given value should be trimmed of trailing and leading whitespace.
        :return: A raw report term for "data_type.field < value".
        """
        data_type: str = AliasUtil.to_data_type_name(data_type) if data_type else self.data_type
        return RawReportTerm(AliasUtil.to_data_type_name(data_type), AliasUtil.to_data_field_name(field), LT,
                             TermBuilder.to_term_val(value), trim)

    def lte_term(self, field: FieldIdentifier, value: TermValue,
                 *, data_type: DataTypeIdentifier | None = None, trim: bool = False) -> RawReportTerm:
        """
        Create a raw report term for comparing a field value with a less than or equal to operation.

        :param field: The data field of this term.
        :param value: The value to compare for this term.
        :param data_type: If provided, the data type of this term. If not provided, then the data type that this term
            builder was instantiated with is used.
        :param trim: Whether the string of the given value should be trimmed of trailing and leading whitespace.
        :return: A raw report term for "data_type.field <= value".
        """
        data_type: str = AliasUtil.to_data_type_name(data_type) if data_type else self.data_type
        return RawReportTerm(AliasUtil.to_data_type_name(data_type), AliasUtil.to_data_field_name(field), LTE,
                             TermBuilder.to_term_val(value), trim)

    def gt_term(self, field: FieldIdentifier, value: TermValue,
                *, data_type: DataTypeIdentifier | None = None, trim: bool = False) -> RawReportTerm:
        """
        Create a raw report term for comparing a field value with a greater than operation.

        :param field: The data field of this term.
        :param value: The value to compare for this term.
        :param data_type: If provided, the data type of this term. If not provided, then the data type that this term
            builder was instantiated with is used.
        :param trim: Whether the string of the given value should be trimmed of trailing and leading whitespace.
        :return: A raw report term for "data_type.field > value".
        """
        data_type: str = AliasUtil.to_data_type_name(data_type) if data_type else self.data_type
        return RawReportTerm(AliasUtil.to_data_type_name(data_type), AliasUtil.to_data_field_name(field), GT,
                             TermBuilder.to_term_val(value), trim)

    def gte_term(self, field: FieldIdentifier, value: TermValue,
                 *, data_type: DataTypeIdentifier | None = None, trim: bool = False) -> RawReportTerm:
        """
        Create a raw report term for comparing a field value with a greater than or equal to operation.

        :param field: The data field of this term.
        :param value: The value to compare for this term.
        :param data_type: If provided, the data type of this term. If not provided, then the data type that this term
            builder was instantiated with is used.
        :param trim: Whether the string of the given value should be trimmed of trailing and leading whitespace.
        :return: A raw report term for "data_type.field >= value".
        """
        data_type: str = AliasUtil.to_data_type_name(data_type) if data_type else self.data_type
        return RawReportTerm(AliasUtil.to_data_type_name(data_type), AliasUtil.to_data_field_name(field), GTE,
                             TermBuilder.to_term_val(value), trim)

    @staticmethod
    def compare_is_term(data_type_A: DataTypeIdentifier, field_A: FieldIdentifier,
                        data_type_B: DataTypeIdentifier, field_B: FieldIdentifier,
                        *, trim: bool = False) -> FieldCompareReportTerm:
        """
        Create a field comparison report term for comparing field values between data types with an equals operation.

        :param data_type_A: The data type for the left side of this term.
        :param field_A: The data field for the left side of this term.
        :param data_type_B: The data type for the right side of this term.
        :param field_B: The data field for the right side of this term.
        :param trim: Whether the field values should be trimmed of trailing and leading whitespace for comparing.
        :return: A field comparison report term for "data_type_A.field_A = data_type_B.field_B".
        """
        return FieldCompareReportTerm(AliasUtil.to_data_type_name(data_type_A), AliasUtil.to_data_field_name(field_A), EQ,
                                      AliasUtil.to_data_type_name(data_type_B), AliasUtil.to_data_field_name(field_B), trim)

    @staticmethod
    def compare_not_term(data_type_A: DataTypeIdentifier, field_A: FieldIdentifier,
                         data_type_B: DataTypeIdentifier, field_B: FieldIdentifier,
                         *, trim: bool = False) -> FieldCompareReportTerm:
        """
        Create a field comparison report term for comparing field values between data types with a not equals operation.

        :param data_type_A: The data type for the left side of this term.
        :param field_A: The data field for the left side of this term.
        :param data_type_B: The data type for the right side of this term.
        :param field_B: The data field for the right side of this term.
        :param trim: Whether the field values should be trimmed of trailing and leading whitespace for comparing.
        :return: A field comparison report term for "data_type_A.field_A != data_type_B.field_B".
        """
        return FieldCompareReportTerm(AliasUtil.to_data_type_name(data_type_A), AliasUtil.to_data_field_name(field_A), NEQ,
                                      AliasUtil.to_data_type_name(data_type_B), AliasUtil.to_data_field_name(field_B), trim)

    @staticmethod
    def compare_lt_term(data_type_A: DataTypeIdentifier, field_A: FieldIdentifier,
                        data_type_B: DataTypeIdentifier, field_B: FieldIdentifier,
                        *, trim: bool = False) -> FieldCompareReportTerm:
        """
        Create a field comparison report term for comparing field values between data types with a less than operation.

        :param data_type_A: The data type for the left side of this term.
        :param field_A: The data field for the left side of this term.
        :param data_type_B: The data type for the right side of this term.
        :param field_B: The data field for the right side of this term.
        :param trim: Whether the field values should be trimmed of trailing and leading whitespace for comparing.
        :return: A field comparison report term for "data_type_A.field_A < data_type_B.field_B".
        """
        return FieldCompareReportTerm(AliasUtil.to_data_type_name(data_type_A), AliasUtil.to_data_field_name(field_A), LT,
                                      AliasUtil.to_data_type_name(data_type_B), AliasUtil.to_data_field_name(field_B), trim)

    @staticmethod
    def compare_lte_term(data_type_A: DataTypeIdentifier, field_A: FieldIdentifier,
                         data_type_B: DataTypeIdentifier, field_B: FieldIdentifier,
                         *, trim: bool = False) -> FieldCompareReportTerm:
        """
        Create a field comparison report term for comparing field values between data types with a less than or equal
        to operation.

        :param data_type_A: The data type for the left side of this term.
        :param field_A: The data field for the left side of this term.
        :param data_type_B: The data type for the right side of this term.
        :param field_B: The data field for the right side of this term.
        :param trim: Whether the field values should be trimmed of trailing and leading whitespace for comparing.
        :return: A field comparison report term for "data_type_A.field_A <= data_type_B.field_B".
        """
        return FieldCompareReportTerm(AliasUtil.to_data_type_name(data_type_A), AliasUtil.to_data_field_name(field_A), LTE,
                                      AliasUtil.to_data_type_name(data_type_B), AliasUtil.to_data_field_name(field_B), trim)

    @staticmethod
    def compare_gt_term(data_type_A: DataTypeIdentifier, field_A: FieldIdentifier,
                        data_type_B: DataTypeIdentifier, field_B: FieldIdentifier,
                        *, trim: bool = False) -> FieldCompareReportTerm:
        """
        Create a field comparison report term for comparing field values between data types with a greater than
        operation.

        :param data_type_A: The data type for the left side of this term.
        :param field_A: The data field for the left side of this term.
        :param data_type_B: The data type for the right side of this term.
        :param field_B: The data field for the right side of this term.
        :param trim: Whether the field values should be trimmed of trailing and leading whitespace for comparing.
        :return: A field comparison report term for "data_type_A.field_A > data_type_B.field_B".
        """
        return FieldCompareReportTerm(AliasUtil.to_data_type_name(data_type_A), AliasUtil.to_data_field_name(field_A), GT,
                                      AliasUtil.to_data_type_name(data_type_B), AliasUtil.to_data_field_name(field_B), trim)

    @staticmethod
    def compare_gte_term(data_type_A: DataTypeIdentifier, field_A: FieldIdentifier,
                         data_type_B: DataTypeIdentifier, field_B: FieldIdentifier,
                         *, trim: bool = False) -> FieldCompareReportTerm:
        """
        Create a field comparison report term for comparing field values between data types with a greater than or
        equal to operation.

        :param data_type_A: The data type for the left side of this term.
        :param field_A: The data field for the left side of this term.
        :param data_type_B: The data type for the right side of this term.
        :param field_B: The data field for the right side of this term.
        :param trim: Whether the field values should be trimmed of trailing and leading whitespace for comparing.
        :return: A field comparison report term for "data_type_A.field_A >= data_type_B.field_B".
        """
        return FieldCompareReportTerm(AliasUtil.to_data_type_name(data_type_A), AliasUtil.to_data_field_name(field_A), GTE,
                                      AliasUtil.to_data_type_name(data_type_B), AliasUtil.to_data_field_name(field_B), trim)

    @staticmethod
    def or_terms(a: AbstractReportTerm, b: AbstractReportTerm, *, is_negated: bool = False) -> CompositeReportTerm:
        """
        Combine two report terms with an OR operation.

        :param a: The first term in the operation.
        :param b: The second term in the operation.
        :param is_negated: Whether the returned term should be negated (i.e. turn this into a nor operation).
        :return: A composite report term for "A or B".
        """
        return CompositeReportTerm(a, OR, b, is_negated)

    @staticmethod
    def and_terms(a: AbstractReportTerm, b: AbstractReportTerm, *, is_negated: bool = False) -> CompositeReportTerm:
        """
        Combine two report terms with an AND operation.

        :param a: The first term in the operation.
        :param b: The second term in the operation.
        :param is_negated: Whether the returned term should be negated (i.e. turn this into a nand operation).
        :return: A composite report term for "A and B".
        """
        return CompositeReportTerm(a, AND, b, is_negated)

    @staticmethod
    def xor_terms(a: AbstractReportTerm, b: AbstractReportTerm, *, is_negated: bool = False) -> CompositeReportTerm:
        """
        Combine two report terms with a XOR operation. Note that a XOR operation doesn't actually exist for custom
        reports. This instead constructs a term that is "(A or B) and !(A and B)", which is equivalent to a XOR
        operation.

        :param a: The first term in the operation.
        :param b: The second term in the operation.
        :param is_negated: Whether the returned term should be negated (i.e. turn this into an xnor operation).
        :return: A composite report term for "A xor B".
        """
        return TermBuilder.and_terms(TermBuilder.or_terms(a, b),
                                     TermBuilder.and_terms(a, b, is_negated=True),
                                     is_negated=is_negated)

    @staticmethod
    def to_term_val(value: TermValue) -> str:
        """
        Convert the given value to be used in a custom report term to a string. Term values may be strings, integers,
        floats, booleans, or lists of values.

        :param value: A value to be used in a custom report term.
        :return: The provided value formatted as a string that can be used
        """
        # If the given value is already a string, then nothing needs to be done with it.
        if not isinstance(value, str):
            # If the given value is None, then use an empty string for the search instead.
            if value is None:
                value = ""
            # If the given value is an iterable object, then the return value is the contents of that iterable
            # in a comma separated list surrounded by curly braces.
            elif isinstance(value, Iterable):
                # When converting a list of values to a string, values in the list which are already strings should be
                # put in quotation marks so that strings that contain commas do not get split up. All other value
                # types can be simply converted to a string, though.
                def convert_list_value(val: TermValue) -> str:
                    return f"'{val}'" if isinstance(val, str) else str(val)
                value = "{" + ",".join([convert_list_value(x) for x in value]) + "}"
            else:
                # Otherwise, the value is simply cast to a string.
                value = str(value)
        return value



from enum import Enum

from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.CustomReport import ReportColumn, CustomReportCriteria

from sapiopycommons.customreport.column_builder import ColumnBuilder
from sapiopycommons.customreport.term_builder import TermBuilder
from sapiopycommons.datatype.pseudo_data_types import AuditLogPseudoDef
from sapiopycommons.general.aliases import RecordIdentifier, AliasUtil, UserIdentifier, FieldIdentifier, FieldValue
from sapiopycommons.general.custom_report_util import CustomReportUtil


class EventType(Enum):
    """An enum to represent the possible event type values with the event type column in the audit log table."""
    ADD = 0
    DELETE = 1
    MODIFY = 2
    INFO = 3
    ERROR = 4
    WARNING = 5
    IMPORT = 6
    GENERATE = 7
    EXPORT = 8
    ADDREF = 9
    REMOVEREF = 10
    ESIGNATURE = 11
    ROLEASSIGNMENT = 12


class AuditLogEntry:
    __event_type: EventType
    __date: int
    __data_type_name: str
    __record_id: int
    __description: str
    __users_login_name: str
    __comment: str
    __data_record_name: str
    __data_field_name: str
    __original_value: str
    __new_value: str

    @property
    def event_type(self) -> EventType:
        return self.__event_type

    @property
    def date(self) -> int:
        return self.__date

    @property
    def data_type_name(self) -> str:
        return self.__data_type_name

    @property
    def record_id(self) -> int:
        return self.__record_id

    @property
    def description(self) -> str:
        return self.__description

    @property
    def users_login_name(self) -> str:
        return self.__users_login_name

    @property
    def comment(self) -> str:
        return self.__comment

    @property
    def data_record_name(self) -> str:
        return self.__data_record_name

    @property
    def data_field_name(self) -> str:
        return self.__data_field_name

    @property
    def original_value(self) -> str:
        return self.__original_value

    @property
    def new_value(self) -> str:
        return self.__new_value

    def __init__(self, report_row: dict[str, FieldValue]):
        self.__event_type = EventType((report_row[AuditLogPseudoDef.EVENT_TYPE__FIELD_NAME.field_name]))
        self.__date = report_row[AuditLogPseudoDef.TIME_STAMP__FIELD_NAME.field_name]
        self.__data_type_name = report_row[AuditLogPseudoDef.DATA_TYPE_NAME__FIELD_NAME.field_name]
        self.__record_id = report_row[AuditLogPseudoDef.RECORD_ID__FIELD_NAME.field_name]
        self.__description = report_row[AuditLogPseudoDef.DESCRIPTION__FIELD_NAME.field_name]
        self.__users_login_name = report_row[AuditLogPseudoDef.USER_NAME__FIELD_NAME.field_name]
        self.__comment = report_row[AuditLogPseudoDef.USER_COMMENT__FIELD_NAME.field_name]
        self.__data_record_name = report_row[AuditLogPseudoDef.RECORD_NAME__FIELD_NAME.field_name]
        self.__data_field_name = report_row[AuditLogPseudoDef.DATA_FIELD_NAME__FIELD_NAME.field_name]
        self.__original_value = report_row[AuditLogPseudoDef.ORIGINAL_VALUE__FIELD_NAME.field_name]
        self.__new_value = report_row[AuditLogPseudoDef.NEW_VALUE__FIELD_NAME.field_name]


class AuditLogUtil:
    user: SapioUser

    def __init__(self, context: UserIdentifier):
        self.user = AliasUtil.to_sapio_user(context)

    @staticmethod
    def report_columns() -> list[ReportColumn]:
        return [
            ColumnBuilder.build_column(AuditLogPseudoDef.DATA_TYPE_NAME, AuditLogPseudoDef.EVENT_TYPE__FIELD_NAME),
            ColumnBuilder.build_column(AuditLogPseudoDef.DATA_TYPE_NAME, AuditLogPseudoDef.TIME_STAMP__FIELD_NAME),
            ColumnBuilder.build_column(AuditLogPseudoDef.DATA_TYPE_NAME, AuditLogPseudoDef.DATA_TYPE_NAME__FIELD_NAME),
            ColumnBuilder.build_column(AuditLogPseudoDef.DATA_TYPE_NAME, AuditLogPseudoDef.RECORD_ID__FIELD_NAME),
            ColumnBuilder.build_column(AuditLogPseudoDef.DATA_TYPE_NAME, AuditLogPseudoDef.DESCRIPTION__FIELD_NAME),
            ColumnBuilder.build_column(AuditLogPseudoDef.DATA_TYPE_NAME, AuditLogPseudoDef.USER_NAME__FIELD_NAME),
            ColumnBuilder.build_column(AuditLogPseudoDef.DATA_TYPE_NAME, AuditLogPseudoDef.USER_COMMENT__FIELD_NAME),
            ColumnBuilder.build_column(AuditLogPseudoDef.DATA_TYPE_NAME, AuditLogPseudoDef.RECORD_NAME__FIELD_NAME),
            ColumnBuilder.build_column(AuditLogPseudoDef.DATA_TYPE_NAME, AuditLogPseudoDef.DATA_FIELD_NAME__FIELD_NAME),
            ColumnBuilder.build_column(AuditLogPseudoDef.DATA_TYPE_NAME, AuditLogPseudoDef.ORIGINAL_VALUE__FIELD_NAME),
            ColumnBuilder.build_column(AuditLogPseudoDef.DATA_TYPE_NAME, AuditLogPseudoDef.NEW_VALUE__FIELD_NAME)
        ]

    @staticmethod
    def create_data_record_audit_log_report(records: list[RecordIdentifier],
                                            fields: list[FieldIdentifier] | None = None) -> CustomReportCriteria:
        """
        This method creates a CustomReportCriteria object for running an audit log query based on data records.

        Creates a CustomReportCriteria object with a query term based on the record ids/records passed into the method.
        Optionally, the fields parameter can be populated to limit the search to particular fields. If the fields
        parameter is not populated, the search will include results for all field changes.

        :param records: The DataRecords, RecordModels, or record ids to base the search on.
        :param fields: The data field names to include changes for.
        :return: The constructed CustomReportCriteria object, which can be used to run a report on the audit log.
        """
        # Build the raw report term querying for any entry with a matching record ID value to the record ID's
        # passed in.
        record_ids = AliasUtil.to_record_ids(records)
        tb = TermBuilder(AuditLogPseudoDef.DATA_TYPE_NAME)
        root_term = tb.is_term(AuditLogPseudoDef.RECORD_ID__FIELD_NAME, record_ids)

        # If the user passed in any specific fields, then we should limit the query to those fields.
        if fields is not None and fields:
            fields: list[str] = AliasUtil.to_data_field_names(fields)
            field_term = tb.is_term(AuditLogPseudoDef.DATA_FIELD_NAME__FIELD_NAME, fields)
            root_term = TermBuilder.and_terms(root_term, field_term)

        return CustomReportCriteria(AuditLogUtil.report_columns(), root_term)

    def run_data_record_audit_log_report(self, records: list[RecordIdentifier],
                                         fields: list[FieldIdentifier] | None = None) \
            -> dict[RecordIdentifier, list[AuditLogEntry]]:
        """
        This method runs a custom report for changes made to the given data records using the audit log.
        See "create_data_record_audit_log_report" for more details about the data record audit log report.

        :param records: The DataRecords, RecordModels, or record ids to base the search on.
        :param fields: The data field names to include changes for.
        :return: A dictionary where the keys are the record identifiers passed in, and the values are a list of
            AuditLogEntry objects which match the record id value of those records.
        """
        # First, we must build our report criteria for running the Custom Report.
        criteria = AuditLogUtil.create_data_record_audit_log_report(records, fields)

        # Then we must run the custom report using that criteria.
        raw_report_data: list[dict[str, FieldValue]] = CustomReportUtil.run_custom_report(self.user, criteria)

        # This section will prepare a map matching the original RecordIdentifier by record id.
        # This is because the audit log entries will have record ids, but we want the keys in our result map
        # to match the record identifiers that the user passed in, for convenience.
        record_identifier_mapping: dict[int, RecordIdentifier] = dict()
        for record in records:
            record_id = AliasUtil.to_record_id(record)
            record_identifier_mapping[record_id] = record

        # Finally, we compile our audit data into a map where the keys are the record identifiers passed in,
        # and the value is a list of applicable audit log entries.
        final_audit_data: dict[RecordIdentifier, list[AuditLogEntry]] = dict()
        for audit_entry_data in raw_report_data:
            audit_entry: AuditLogEntry = AuditLogEntry(audit_entry_data)
            identifier: RecordIdentifier = record_identifier_mapping.get(audit_entry.record_id)
            final_audit_data.setdefault(identifier, []).append(audit_entry)

        return final_audit_data



import base64
from enum import Enum

from databind.core.dataclasses import dataclass


class ChannelStatisticType(Enum):
    """
    All supported channel statistics type.
    """
    MEAN = "(Mean) MFI"
    MEDIAN = "(Median) MFI"
    STD_EV = "Std. Dev."
    COEFFICIENT_OF_VARIATION = "CV"

    display_name: str

    def __init__(self, display_name: str):
        self.display_name = display_name


@dataclass
class ChannelStatisticsParameterJSON:
    channelNameList: list[str]
    statisticsType: ChannelStatisticType

    def __init__(self, channel_name_list: list[str], stat_type: ChannelStatisticType):
        self.channelNameList = channel_name_list
        self.statisticsType = stat_type


@dataclass
class ComputeFlowStatisticsInputJson:
    fcsFileRecordIdList: list[int]
    statisticsParameterList: list[ChannelStatisticsParameterJSON]

    def __init__(self, fcs_file_record_id_list: list[int], statistics_parameter_list: list[ChannelStatisticsParameterJSON]):
        self.fcsFileRecordIdList = fcs_file_record_id_list
        self.statisticsParameterList = statistics_parameter_list


@dataclass
class FlowJoWorkspaceInputJson:
    filePath: str
    base64Data: str

    def __init__(self, filePath: str, file_data: bytes):
        self.filePath = filePath
        self.base64Data = base64.b64encode(file_data).decode('utf-8')


@dataclass
class UploadFCSInputJson:
    """
    Request to upload new FCS file
    Attributes:
        filePath: The file name of the FCS file to be uploaded. For FlowJo workspace, this is important to match the file in group (via file names).
        attachmentDataType: the attachment data type that contains already-uploaded FCS data.
        attachmentRecordId: the attachment record ID that contains already-uploaded FCS data.
        associatedRecordDataType: the "parent" association for the FCS. Can either be a workspace or a sample record.
        associatedRecordId: the "parent" association for the FCS. Can either be a workspace or a sample record.
    """
    filePath: str
    attachmentDataType: str
    attachmentRecordId: int
    associatedRecordDataType: str
    associatedRecordId: int

    def __init__(self, associated_record_data_type: str, associated_record_id: int,
                 file_path: str, attachment_data_type: str, attachment_record_id: int):
        self.filePath = file_path
        self.attachmentDataType = attachment_data_type
        self.attachmentRecordId = attachment_record_id
        self.associatedRecordDataType = associated_record_data_type
        self.associatedRecordId = associated_record_id



from __future__ import annotations

from weakref import WeakValueDictionary

from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.DataRecord import DataRecord
from sapiopylib.rest.pojo.eln.SapioELNEnums import ElnBaseDataType
from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
from sapiopylib.rest.utils.recordmodel.RecordModelManager import RecordModelManager, RecordModelInstanceManager
from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrappedType

from sapiopycommons.general.aliases import FieldMap, AliasUtil, DataTypeIdentifier
from sapiopycommons.general.exceptions import SapioException


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
class ElnRuleHandler:
    """
    A class which helps with the parsing and navigation of the ELN rule result map of a webhook context.
    """
    __context: SapioWebhookContext
    """The context that this handler is working from."""

    __inst_man: RecordModelInstanceManager
    """The record model instance manager, used for wrapping the data records as record models."""

    # Reformatted and cached version of the Velox ELN rule result map for easier handling.
    __records: dict[str, set[DataRecord]]
    """A mapping of data type to the set of data records from the context that match that data type."""
    __entry_to_records: dict[str, dict[str, set[DataRecord]]]
    """A mapping of entry name to the sets of data records for that entry, each set of records being mapped by its
    data type."""
    __field_maps: dict[str, dict[int, FieldMap]]
    """A mapping of data type to the field maps from the context that match that data type. In order to prevent
    duplicate field maps, each field map is in a dict keyed by the RecordId field in the field map, since field maps
    are just dictionaries and dictionaries aren't hashable and therefore can't go in a set."""
    __entry_to_field_maps: dict[str, dict[str, dict[int, FieldMap]]]
    """A mapping of entry name to the lists of field maps for that entry, each grouping of field maps being mapped by
    its data type."""

    __instances: WeakValueDictionary[SapioUser, ElnRuleHandler] = WeakValueDictionary()
    __initialized: bool

    def __new__(cls, context: SapioWebhookContext):
        if context.velox_eln_rule_result_map is None:
            raise SapioException("No Velox ELN rule result map in context for ElnRuleHandler to parse.")
        user = context if isinstance(context, SapioUser) else context.user
        obj = cls.__instances.get(user)
        if not obj:
            obj = object.__new__(cls)
            obj.__initialized = False
            cls.__instances[user] = obj
        return obj

    def __init__(self, context: SapioWebhookContext):
        if self.__initialized:
            return
        self.__initialized = True

        if context.velox_eln_rule_result_map is None:
            raise SapioException("No Velox ELN rule result map in context for ElnRuleHandler to parse.")
        self.__context = context
        self.__inst_man = RecordModelManager(context.user).instance_manager
        self.__cache_records()

    def __cache_records(self) -> None:
        """
        Cache the records from the context into dictionaries. Two caches are created. One cache maps the data type of
        each record to a set of all records of that data type. The other cache maps the entry that the records come
        from to another dict that maps the data types to the records of that type.
        Doesn't cache any relationship info from the VeloxRuleType of the rule results.
        """
        self.__records = {}
        self.__entry_to_records = {}
        # Each entry in the context has a list of results for that entry.
        for entry, entry_results in self.__context.velox_eln_rule_result_map.items():
            # Keep track of the records for this specific entry.
            entry_dict: dict[str, set[DataRecord]] = {}
            # Entry results consist of a record ID of the record in the entry and a list of results tied to that record.
            for record_result in entry_results:
                # The list of results for a record consist of a list of data records and a VeloxType that specifies
                # how the records in the list relate to the main record.
                for result in record_result.rule_result_list:
                    # For the purposes of caching, we don't care about the VeloxType.
                    for record in result.data_records:
                        # Get the data type of this record. If this is an ELN type, ignore the digits.
                        data_type: str = record.data_type_name
                        # PR-46331: Ensure that all ELN types are converted to their base data type name.
                        if ElnBaseDataType.is_eln_type(data_type):
                            data_type = ElnBaseDataType.get_base_type(data_type).data_type_name
                        # Update the list of records of this type that exist so far globally.
                        self.__records.setdefault(data_type, set()).add(record)
                        # Do the same for the list of records of this type for this specific entry.
                        entry_dict.setdefault(data_type, set()).add(record)
            # Update the records for this entry.
            self.__entry_to_records.update({entry: entry_dict})

        self.__field_maps = {}
        self.__entry_to_field_maps = {}
        # Repeat the same thing for the field map results.
        for entry, entry_results in self.__context.velox_eln_rule_field_map_result_map.items():
            entry_dict: dict[str, dict[int, FieldMap]] = {}
            for record_result in entry_results:
                for result in record_result.velox_type_rule_field_map_result_list:
                    data_type: str = result.velox_type_pojo.data_type_name
                    if ElnBaseDataType.is_eln_type(data_type):
                        data_type = ElnBaseDataType.get_base_type(data_type).data_type_name
                    for field_map in result.field_map_list:
                        rec_id: int = field_map.get("RecordId")
                        self.__field_maps.setdefault(data_type, {}).update({rec_id: field_map})
                        entry_dict.setdefault(data_type, {}).update({rec_id: field_map})
            self.__entry_to_field_maps.update({entry: entry_dict})

    def get_entry_names(self) -> list[str]:
        """
        :return: A list of the entry names that may be used with the get_records and get_models functions. These are the
            entries from the experiment that the records in the rule context originate from.
        """
        return list(self.__entry_to_records.keys())

    def get_field_maps_entry_names(self) -> list[str]:
        """
        :return: A list of the entry names that may be used with the get_field_maps function. These are the
            entries from the experiment that the field maps in the rule context originate from.
        """
        return list(self.__entry_to_field_maps.keys())

    def get_records(self, data_type: DataTypeIdentifier, entry: str | None = None) -> list[DataRecord]:
        """
        Get records from the cached context with the given data type. Capable of being filtered to searching within
        the context of an entry name. If the given data type or entry does not exist in the context,
        returns an empty list.

        :param data_type: The data type of the records to return.
        :param entry: The name of the entry to grab the records from. If None, returns the records that match the data
            type from every entry. If an entry is provided, but it does not exist in the context, returns an empty list.
        :return: The records from the context that match the input parameters.
        """
        data_type: str = AliasUtil.to_data_type_name(data_type)
        records: dict[str, set[DataRecord]] = self.__entry_to_records.get(entry, {}) if entry else self.__records
        return list(records.get(data_type, []))

    # FR-46701: Add functions to the rule handlers for accessing the field maps of inaccessible records in the context.
    def get_field_maps(self, data_type: DataTypeIdentifier, entry: str | None = None) -> list[FieldMap]:
        """
        Get field maps from the cached context with the given data type. Capable of being filtered to searching within
        the context of an entry name. If the given data type or entry does not exist in the context,
        returns an empty list.

        Field maps will only exist in the context if the data record that the fields are from is no longer accessible
        to the user. This can occur because the data record was deleted, or because the user does not have access to the
        record due to ACL.

        :param data_type: The data type of the field maps to return.
        :param entry: The name of the entry to grab the field maps from. If None, returns the field maps that match the
            data type from every entry. If an entry is provided, but it does not exist in the context, returns an empty
            list.
        :return: The field maps from the context that match the input parameters.
        """
        data_type: str = AliasUtil.to_data_type_name(data_type)
        field_maps: dict[str, dict[int, FieldMap]] = self.__entry_to_field_maps.get(entry, {}) if entry else self.__field_maps
        return list(field_maps.get(data_type, {}).values())

    def get_models(self, wrapper_type: type[WrappedType], entry: str | None = None) -> list[WrappedType]:
        """
        Get record models from the cached context with the given data type. Capable of being filtered to searching
        within the context of an entry name. If the given data type or entry does not exist in the context,
        returns an empty list.

        :param wrapper_type: The record model wrapper to use.
        :param entry: The name of the entry to grab the records from. If None, returns the records that match the data
            type from every entry. If an entry is provided, but it does not exist in the context, returns an empty list.
        :return: The record models from the context that match the input parameters.
        """
        dt: str = wrapper_type.get_wrapper_data_type_name()
        return self.__inst_man.add_existing_records_of_type(self.get_records(dt, entry), wrapper_type)



from sapiopylib.rest.pojo.CustomReport import ReportColumn
from sapiopylib.rest.pojo.datatype.FieldDefinition import FieldType

from sapiopycommons.general.aliases import DataTypeIdentifier, FieldIdentifier, AliasUtil
from sapiopycommons.general.exceptions import SapioException

# The system fields that every record has and their field types. System fields aren't generated as record model fields
# for all platform version, hence the need to create a dict for them in the off chance that they're not present on
# the model wrapper.
SYSTEM_FIELDS: dict[str, FieldType] = {
    "DataRecordName": FieldType.IDENTIFIER,
    "RecordId": FieldType.LONG,
    "DateCreated": FieldType.DATE,
    "CreatedBy": FieldType.STRING,
    "VeloxLastModifiedDate": FieldType.DATE,
    "VeloxLastModifiedBy": FieldType.STRING
}


class ColumnBuilder:
    """
    A class for building report columns for custom reports.
    """
    @staticmethod
    def build_column(data_type: DataTypeIdentifier, field: FieldIdentifier, field_type: FieldType | None = None) \
            -> ReportColumn:
        """
        Build a ReportColumn from a variety of possible inputs.

        :param data_type: An object that can be used to identify a data type.
        :param field: An object that can be used to identify a data field.
        :param field_type: The field type of the provided field. This is only required if the field type cannot be
            determined from the given data type and field, which occurs when the given field is a string and the
            given data type is not a wrapped record model or record model wrapper.
        :return: A ReportColumn for the inputs.
        """
        # Get the data type and field names from the inputs.
        data_type_name = AliasUtil.to_data_type_name(data_type)
        field_name = AliasUtil.to_data_field_name(field)
        if field_type is None:
            field_type = ColumnBuilder.__field_type(data_type, field)
        if field_type is None:
            raise SapioException("The field_type parameter is required for the provided data_type and field inputs.")
        return ReportColumn(data_type_name, field_name, field_type)

    @staticmethod
    def __field_type(data_type: DataTypeIdentifier, field: FieldIdentifier) -> FieldType | None:
        """
        Given a record model wrapper and a field name, return the field type for that field. Accounts for system fields.

        :param data_type: The record model wrapper that the field is on.
        :param field: The field name to return the type of.
        :return: The field type of the given field name.
        """
        # Check if the field name is a system field. If it is, use the field type defined in this file.
        field_name: str = AliasUtil.to_data_field_name(field)
        if field_name in SYSTEM_FIELDS:
            return SYSTEM_FIELDS.get(field_name)
        # Otherwise, check if the field type can be found from the wrapper.
        return AliasUtil.to_field_type(field, data_type)



from collections.abc import Iterable
from typing import Any, TypeAlias

from sapiopylib.rest.User import SapioUser
from sapiopylib.rest.pojo.DataRecord import DataRecord
from sapiopylib.rest.pojo.datatype.FieldDefinition import FieldType, AbstractVeloxFieldDefinition
from sapiopylib.rest.pojo.eln.ElnExperiment import ElnExperiment
from sapiopylib.rest.pojo.eln.ExperimentEntry import ExperimentEntry
from sapiopylib.rest.pojo.eln.SapioELNEnums import ElnBaseDataType
from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
from sapiopylib.rest.utils.Protocols import ElnExperimentProtocol, ElnEntryStep
from sapiopylib.rest.utils.recordmodel.PyRecordModel import PyRecordModel, AbstractRecordModel
from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrappedRecordModel, WrappedType, WrapperField

from sapiopycommons.general.exceptions import SapioException

FieldValue: TypeAlias = int | float | str | bool | None
"""Allowable values for fields in the system."""
RecordModel: TypeAlias = PyRecordModel | AbstractRecordModel | WrappedRecordModel
"""Different forms that a record model could take."""
SapioRecord: TypeAlias = DataRecord | RecordModel
"""A record could be provided as either a DataRecord, PyRecordModel, or WrappedRecordModel (WrappedType)."""
RecordIdentifier: TypeAlias = SapioRecord | int
"""A RecordIdentifier is either a record type or an integer for the record's record ID."""
DataTypeIdentifier: TypeAlias = SapioRecord | type[WrappedType] | str
"""A DataTypeIdentifier is either a SapioRecord, a record model wrapper type, or a string."""
FieldIdentifier: TypeAlias = AbstractVeloxFieldDefinition | WrapperField | str | tuple[str, FieldType]
"""A FieldIdentifier is either wrapper field from a record model wrapper, a string, or a tuple of string
and field type."""
FieldIdentifierKey: TypeAlias = WrapperField | str
"""A FieldIdentifierKey is a FieldIdentifier, except it can't be a tuple, s tuples can't be used as keys in
dictionaries.."""
HasFieldWrappers: TypeAlias = type[WrappedType] | WrappedRecordModel
"""An identifier for classes that have wrapper fields."""
ExperimentIdentifier: TypeAlias = ElnExperimentProtocol | ElnExperiment | int
"""An ExperimentIdentifier is either an experiment protocol, experiment, or an integer for the experiment's notebook
ID."""
ExperimentEntryIdentifier: TypeAlias = ElnEntryStep | ExperimentEntry | int
"""An ExperimentEntryIdentifier is either an ELN entry step, experiment entry, or an integer for the entry's ID."""
FieldMap: TypeAlias = dict[str, FieldValue]
"""A field map is simply a dict of data field names to values. The purpose of aliasing this is to help distinguish
any random dict in a webhook from one which is explicitly used for record fields."""
FieldIdentifierMap: TypeAlias = dict[FieldIdentifierKey, FieldValue]
"""A field identifier map is the same thing as a field map, except the keys can be field identifiers instead
of just strings. Note that although one of the allowed field identifiers is a tuple, you can't use tuples as
keys in a dictionary."""
UserIdentifier: TypeAlias = SapioWebhookContext | SapioUser
"""An identifier for classes from which a user object can be used for sending requests."""


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
class AliasUtil:
    @staticmethod
    def to_data_record(record: SapioRecord) -> DataRecord:
        """
        Convert a single DataRecord, PyRecordModel, or WrappedRecordModel to just a DataRecord.

        :return: The DataRecord of the input SapioRecord.
        """
        return record if isinstance(record, DataRecord) else record.get_data_record()

    @staticmethod
    def to_data_records(records: Iterable[SapioRecord]) -> list[DataRecord]:
        """
        Convert a list of variables that could either be DataRecords, PyRecordModels,
        or WrappedRecordModels to just DataRecords.

        :return: A list of DataRecords for the input records.
        """
        return [(x if isinstance(x, DataRecord) else x.get_data_record()) for x in records]

    @staticmethod
    def to_record_ids(records: Iterable[RecordIdentifier]) -> list[int]:
        """
        Convert a list of variables that could either be integers, DataRecords, PyRecordModels,
        or WrappedRecordModels to just integers (taking the record ID from the records).

        :return: A list of record IDs for the input records.
        """
        return [(AliasUtil.to_record_id(x)) for x in records]

    @staticmethod
    def to_record_id(record: RecordIdentifier):
        """
        Convert a single variable that could be either an integer, DataRecord, PyRecordModel,
        or WrappedRecordModel to just an integer (taking the record ID from the record).

        :return: A record ID for the input record.
        """
        return record if isinstance(record, int) else record.record_id

    @staticmethod
    def to_data_type_name(value: DataTypeIdentifier, convert_eln_dts: bool = True) -> str:
        """
        Convert a given value to a data type name.

        :param value: A value which is a string, record, or record model type.
        :param convert_eln_dts: If true, convert ELN data types to their base data type name.
        :return: A string of the data type name of the input value.
        """
        if isinstance(value, SapioRecord):
            value = value.data_type_name
        elif not isinstance(value, str):
            value = value.get_wrapper_data_type_name()
        if convert_eln_dts and ElnBaseDataType.is_eln_type(value):
            return ElnBaseDataType.get_base_type(value).data_type_name
        return value

    @staticmethod
    def to_data_type_names(values: Iterable[DataTypeIdentifier], return_set: bool = False,
                           convert_eln_dts: bool = True) -> list[str] | set[str]:
        """
        Convert a given iterable of values to a list or set of data type names.

        :param values: An iterable of values which are strings, records, or record model types.
        :param return_set: If true, return a set instead of a list.
        :param convert_eln_dts: If true, convert ELN data types to their base data type name.
        :return: A list or set of strings of the data type name of the input value.
        """
        values = [AliasUtil.to_data_type_name(x, convert_eln_dts) for x in values]
        return set(values) if return_set else values

    @staticmethod
    def to_singular_data_type_name(values: Iterable[DataTypeIdentifier], convert_eln_dts: bool = True) -> str:
        """
        Convert a given iterable of values to a singular data type name that they share. Throws an exception if more
        than one data type name exists in the provided list of identifiers.

        :param values: An iterable of values which are strings, records, or record model types.
        :param convert_eln_dts: If true, convert ELN data types to their base data type name.
        :return: The single data type name that the input vales share. Returns an empty string if an empty iterable
            was provided.
        """
        if not values:
            return ""
        data_types: set[str] = AliasUtil.to_data_type_names(values, True, convert_eln_dts)
        if len(data_types) > 1:
            raise SapioException(f"Provided values contain multiple data types: {data_types}. "
                                 f"Only expecting a single data type.")
        return data_types.pop()

    @staticmethod
    def to_data_field_name(value: FieldIdentifier) -> str:
        """
        Convert an object that can be used to identify a data field to a data field name string.

        :param value: An object that can be used to identify a data field.
        :return: A string of the data field name of the input value.
        """
        if isinstance(value, tuple):
            return value[0]
        if isinstance(value, WrapperField):
            return value.field_name
        if isinstance(value, AbstractVeloxFieldDefinition):
            return value.data_field_name
        return value

    @staticmethod
    def to_data_field_names(values: Iterable[FieldIdentifier]) -> list[str]:
        """
        Convert an iterable of objects that can be used to identify data fields to a list of data field name strings.

        :param values: An iterable of objects that can be used to identify a data field.
        :return: A list of strings of the data field names of the input values.
        """
        return [AliasUtil.to_data_field_name(x) for x in values]

    @staticmethod
    def to_data_field_names_dict(values: dict[FieldIdentifierKey, Any]) -> dict[str, Any]:
        """
        Take a dictionary whose keys are field identifiers and convert them all to strings for the data field name.

        :param values: A dictionary of field identifiers to field values.
        :return: A dictionary of strings of the data field names to field values for the input values.
        """
        ret_dict: dict[str, FieldValue] = {}
        for field, value in values.items():
            ret_dict[AliasUtil.to_data_field_name(field)] = value
        return ret_dict

    @staticmethod
    def to_data_field_names_list_dict(values: list[dict[FieldIdentifierKey, Any]]) -> list[dict[str, Any]]:
        ret_list: list[dict[str, Any]] = []
        for field_map in values:
            ret_list.append(AliasUtil.to_data_field_names_dict(field_map))
        return ret_list

    @staticmethod
    def to_field_type(field: FieldIdentifier, data_type: HasFieldWrappers | None = None) -> FieldType:
        """
        Convert a given field identifier to the field type for that field.

        :param field: A string or WrapperField.
        :param data_type: If the field is provided as a string, then a record model wrapper or wrapped record model
            must be provided to determine the field type.
        :return: The field type of the given field.
        """
        if isinstance(field, tuple):
            return field[1]
        if isinstance(field, WrapperField):
            return field.field_type
        for var in dir(data_type):
            attr = getattr(data_type, var)
            if isinstance(attr, WrapperField) and attr.field_name == field:
                return attr.field_type
        raise SapioException(f"The wrapper of data type \"{data_type.get_wrapper_data_type_name()}\" doesn't have a "
                             f"field with the name \"{field}\",")

    @staticmethod
    def to_field_map(record: SapioRecord) -> FieldMap:
        """
        Convert a given record value to a field map. This includes the given RecordId of the given record.

        :return: The field map for the input record.
        """
        if isinstance(record, DataRecord):
            # noinspection PyTypeChecker
            fields: FieldMap = record.get_fields()
        else:
            fields: FieldMap = record.fields.copy_to_dict()
        fields["RecordId"] = AliasUtil.to_record_id(record)
        return fields

    @staticmethod
    def to_field_map_lists(records: Iterable[SapioRecord]) -> list[FieldMap]:
        """
        Convert a list of variables that could either be DataRecords, PyRecordModels, or WrappedRecordModels
        to a list of their field maps. This includes the given RecordId of the given records.

        :return: A list of field maps for the input records.
        """
        field_map_list: list[FieldMap] = []
        for record in records:
            field_map_list.append(AliasUtil.to_field_map(record))
        return field_map_list

    @staticmethod
    def to_notebook_id(experiment: ExperimentIdentifier) -> int:
        """
        Convert an object that identifies an ELN experiment to its notebook ID.

        :return: The notebook ID for the experiment identifier.
        """
        if isinstance(experiment, int):
            return experiment
        if isinstance(experiment, ElnExperiment):
            return experiment.notebook_experiment_id
        return experiment.get_id()

    @staticmethod
    def to_notebook_ids(experiments: list[ExperimentIdentifier]) -> list[int]:
        """
        Convert a list of objects that identify ELN experiments to their notebook IDs.

        :return: The list of notebook IDs for the experiment identifiers.
        """
        notebook_ids: list[int] = []
        for experiment in experiments:
            notebook_ids.append(AliasUtil.to_notebook_id(experiment))
        return notebook_ids

    @staticmethod
    def to_entry_id(entry: ExperimentEntryIdentifier) -> int:
        """
        Convert an object that identifies an experiment entry to its entry ID.

        :return: The entry ID for the entry identifier.
        """
        if isinstance(entry, int):
            return entry
        elif isinstance(entry, ExperimentEntry):
            return entry.entry_id
        elif isinstance(entry, ElnEntryStep):
            return entry.get_id()
        raise SapioException(f"Unrecognized entry identifier of type {type(entry)}")

    @staticmethod
    def to_entry_ids(entries: list[ExperimentEntryIdentifier]) -> list[int]:
        """
        Convert a list of objects that identify experiment entries to their entry IDs.

        :return: The list of entry IDs for the entry identifiers.
        """
        entry_ids: list[int] = []
        for entry in entries:
            entry_ids.append(AliasUtil.to_entry_id(entry))
        return entry_ids

    @staticmethod
    def to_sapio_user(context: UserIdentifier) -> SapioUser:
        """
        Convert an object that could be either a SapioUser or SapioWebhookContext to just a SapioUser.

        :return: A SapioUser object.
        """
        return context if isinstance(context, SapioUser) else context.user



import unittest

from sapiopylib.rest.User import SapioUser

from sapiopycommons.multimodal.multimodal import MultiModalManager
from sapiopycommons.multimodal.multimodal_data import MultiSequenceAlignmentRequestPojo, MAFFTRunOptions, MAFFTStrategy, \
    MultiSequenceAlignmentTool, MultiSequenceAlignmentSeqPojo, MultiSeqAlignemntSeqType

user = SapioUser(url="https://linux-vm:8443/webservice/api", verify_ssl_cert=False,
                 guid="3c232543-f407-4828-aae5-b33d4cd31fa7",
                 username="yqiao_api", password="Password1!")


class MAFFT_Test(unittest.TestCase):
    def test_plain_json(self):
        json = """
{
  "inputSeqs": [
    {
      "seqId": "1ed1b624-e687-46a4-809f-3bd4b6c092b6",
      "seqString": "MMFSGFNADYEASSSRCSSASPAGDSLSYYHSPADSFSSMGSPVNAQDFCTDLAVSSANFIPTVTAISTSPDLQWLVQPALVSSVAPSQTRAPHPFGVPAPSAGAYSRAGVVKTMTGGRAQSIGRRGKVEQLSPEEEEKRRIRRERNKMAAAKCRNRRRELTDTLQAETDQLEDEKSALQTEIANLLKEKEKLEFILAAHRPACKIPDDLGFPEEMSVASLDLTGGLPEVATPESEEAFTLPLLNDPEPKPSVEPVKSISSMELKTEPFDDFLFPASSRPSGSETARSVPDMDLSGSFYAADWEPLHSGSLGMGPMATELEPLCTPVVTCTPSCTAYTSSFVFTYPEADSFPSCAAAHRKGSSSNEPSSDSLSSPTLLAL"
    },
    {
      "seqId": "63a5827e-68d0-4cb8-a60b-54837eb872d5",
      "seqString": "MFQAFPGDYDSGSRCSSSPSAESQYLSSVDSFGSPPTAAASQECAGLGEMPGSFVPTVTAITTSQDLQWLVQPTLISSMAQSQGQPLASQPPVVDPYDMPGTSYSTPGMSGYSSGGASGSGGPSTSGTTSGPGPARPARARPRRPREETETDQLEEEKAELESEIAELQKEKERLEFVLVAHKPGCKIPYEEGPGPGPLAEVRDLPGSAPAKEDGFSWLLPPPPPPPLPFQTSQDAPPNLTASLFTHSEVQVLGDPFPVVNPSYTSSFVLTCPEVSAFAGAQRTSGSDQPSDPLNSPSLLAL"
    },
    {
      "seqId": "9f64a046-a0ce-4e5c-87b3-fc2fcbae5e4d",
      "seqString": "MFRDFGEPGPSSGNGGGYGGPAQPPAAAQAAQQISPEEEERRRVRRERNKLAAAKCRNRRKELTDFLQAETDKLEDEKSGLQREIEELQKQKERLELVLEAHRPICKIPEGAKEGDTGSTSGTSSPPAPCRPVPCISLSPGPVLEPEALHTPTLMTTPSLTPFTPSLVFTYPSTPEPCASAHRKSSSSSGDPSSDPLGSPTLLAL"
    },
    {
      "seqId": "0f829b72-e35b-4018-be02-7413d56d8309",
      "seqString": "MYQDYPGNFDTSSRGSSGSPAHAESYSSGGGGQQKFRVDMPGSGSAFIPTINAITTSQDLQWMVQPTVITSMSNPYPRSHPYSPLPGLASVPGHMALPRPGVIKTIGTTVGRRRRDEQLSPEEEEKRRIRRERNKLAAAKCRNRRRELTEKLQAETEELEEEKSGLQKEIAELQKEKEKLEFMLVAHGPVCKISPEERRSPPAPGLQPMRSGGGSVGAVVVKQEPLEEDSPSSSSAGLDKAQRSVIKPISIAGGFYGEEPLHTPIVVTSTPAVTPGTSNLVFTYPSVLEQESPASPSESCSKAHRRSSSSGDQSSDSLNSPTLLAL"
    }
  ],
  "parameters": {
    "strategy": "L_INS_i",
    "noScore": true,
    "op": 1.53,
    "ep": 0.123,
    "lop": -2,
    "lep": 0.1,
    "lexp": -0.1,
    "LOP": -6,
    "LEXP": 0,
    "fmodel": false
  },
  "tool": "MAFFT",
  "seqType": "amino"
}
        """
        response = user.plugin_post("bio/multisequencealignment", payload=json, is_payload_plain_text=True)
        user.raise_for_status(response)
        self.assertTrue(response.text)
        print(response.text)

    def test_multimodal_data(self):
        from  sapiopycommons.multimodal.multimodal_data import MultiSequenceAlignmentSeqPojo as SeqPojo
        options = MAFFTRunOptions(strategy=MAFFTStrategy.FFT_NS_2, op=1.6).to_json()
        sequences = [SeqPojo.create("1ed1b624-e687-46a4-809f-3bd4b6c092b6", "MMFSGFNADYEASSSRCSSASPAGDSLSYYHSPADSFSSMGSPVNAQDFCTDLAVSSANFIPTVTAISTSPDLQWLVQPALVSSVAPSQTRAPHPFGVPAPSAGAYSRAGVVKTMTGGRAQSIGRRGKVEQLSPEEEEKRRIRRERNKMAAAKCRNRRRELTDTLQAETDQLEDEKSALQTEIANLLKEKEKLEFILAAHRPACKIPDDLGFPEEMSVASLDLTGGLPEVATPESEEAFTLPLLNDPEPKPSVEPVKSISSMELKTEPFDDFLFPASSRPSGSETARSVPDMDLSGSFYAADWEPLHSGSLGMGPMATELEPLCTPVVTCTPSCTAYTSSFVFTYPEADSFPSCAAAHRKGSSSNEPSSDSLSSPTLLAL"),
                     SeqPojo.create("0f829b72-e35b-4018-be02-7413d56d8309", "MYQDYPGNFDTSSRGSSGSPAHAESYSSGGGGQQKFRVDMPGSGSAFIPTINAITTSQDLQWMVQPTVITSMSNPYPRSHPYSPLPGLASVPGHMALPRPGVIKTIGTTVGRRRRDEQLSPEEEEKRRIRRERNKLAAAKCRNRRRELTEKLQAETEELEEEKSGLQKEIAELQKEKEKLEFMLVAHGPVCKISPEERRSPPAPGLQPMRSGGGSVGAVVVKQEPLEEDSPSSSSAGLDKAQRSVIKPISIAGGFYGEEPLHTPIVVTSTPAVTPGTSNLVFTYPSVLEQESPASPSESCSKAHRRSSSSGDQSSDSLNSPTLLAL")]
        request = MultiSequenceAlignmentRequestPojo(MultiSequenceAlignmentTool.MAFFT, MultiSeqAlignemntSeqType.amino, sequences, options)
        reg_man = MultiModalManager(user)
        result = reg_man.run_multi_sequence_alignment(request)
        self.assertTrue(len(result) == 2)
        print(result)



from enum import Enum


class MessageDisplayType(Enum):
    """
    An enum representing the different ways in which a message can be displayed to the user.
    """
    TOASTER_SUCCESS = 0
    TOASTER_INFO = 1
    TOASTER_WARNING = 2
    TOASTER_ERROR = 3
    OK_DIALOG = 4
    DISPLAY_INFO = 5
    DISPLAY_WARNING = 6
    DISPLAY_ERROR = 7


# FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
class SapioException(Exception):
    """
    A generic exception thrown by sapiopycommons methods. Typically caused by programmer error, but may also be from
    extremely edge case user errors. For expected user errors, use SapioUserErrorException.

    CommonsWebhookHandler's default behavior for this and any other exception that doesn't extend SapioException is
    to return a generic toaster message saying that an unexpected error has occurred.
    """
    pass


class SapioUserCancelledException(SapioException):
    """
    An exception thrown when the user cancels a client callback.

    CommonsWebhookHandler's default behavior is to simply end the webhook session with a true result without logging
    the exception.
    """
    pass


class SapioDialogTimeoutException(SapioException):
    """
    An exception thrown when the user leaves a client callback open for too long.

    CommonsWebhookHandler's default behavior is to display an OK popup notifying the user that the dialog has timed out.
    """
    pass


class DisplayableException(SapioException):
    """
    A generic exception that promises to return a user-friendly message explaining the error that should be displayed to
    the user. Note that it is up to whichever class that catches this exception to actually display the message.
    """
    msg: str
    display_type: MessageDisplayType | None
    title: str | None

    def __init__(self, msg: str, display_type: MessageDisplayType | None = None, title: str | None = None):
        """
        :param msg: The message that should be displayed to the user.
        :param display_type: The manner in which the message should be displayed. If None, then the display type should
            be controlled by the class that catches this exception.
        :param title: If the display type is able to have a title, this is the title that will be displayed. If None,
            then the title should be controlled by the class that catches this exception.
        """
        self.msg = msg
        self.display_type = display_type
        self.title = title


class SapioUserErrorException(DisplayableException):
    """
    An exception caused by user error (e.g. user provided a CSV when an XLSX was expected), which promises to return a
    user-friendly message explaining the error that should be displayed to the user.

    CommonsWebhookHandler's default behavior is to return the error message in a toaster popup.
    """
    pass


class SapioCriticalErrorException(DisplayableException):
    """
    A critical exception caused by user error, which promises to return a user-friendly message explaining the error
    that should be displayed to the user.

    CommonsWebhookHandler's default behavior is to return the error message in a display_error callback.
    """
    pass



# Includes general Multimodal registration data structures for specific endpoints.
# Author: yqiao
import base64
from enum import Enum
from typing import Any

from databind.core import ExtraKeys
from databind.core.dataclasses import dataclass


@dataclass
class PySimpleMoleculePojo:
    imageSVG: str | None
    smiles: str
    molBlock: str


@dataclass
@ExtraKeys()
class PyMolecule:
    """
    Describes a deserialized molecule for Java.
    """
    imageSVG: str | None
    smiles: str
    clogP: float
    tpsa: float
    amw: float
    exactMass: float
    numHBondDonors: int
    numHBondAcceptors: int
    molFormula: str
    charge: int
    molBlock: str
    inchi: str
    inchiKey: str
    stereoisomers: list[PySimpleMoleculePojo] | None
    normError: str | None
    desaltError: str | None
    desaltedList: list[str] | None
    registrationHash: str | None
    hasOrGroup: bool
    CXSMILESHash: str | None


@dataclass
class PyCompound:
    """
    class PyCompound
    """

    # The original substance data.
    originalMol: PyMolecule  # since May 24, 2023

    # The normalized, desalted abstract compound data.
    canonicalMol: PyMolecule | None  # since May 24, 2023

    props: dict[str, object] | None


class ChemFileType(Enum):
    CSV = 0
    SDF = 1


class ChemLoadType(Enum):
    SMILES_LIST = 0
    INCHI_LIST = 1
    MOL_BLOCK_LIST = 2
    SDF_FILE = 3


class ChemSearchType(Enum):
    """
    All possible chemistry quicksearch types.
    """
    COMPOUND_SUBSTRUCTURE = 0
    COMPOUND_SIMILARITY = 1
    REACTION_SUBSTRUCTURE = 2


@dataclass
class ChemLoadingError:
    """
    Describes a single record of compound loading error, this will be produced by interactive loader of compounds.
    Attributes:
        original: The original data that makes up the molecule so user can find it in the file.
        errorMsg: Why the import has failed. (normalization/desalting/sanitization/kekulize)
        properties: The additional attributes in the import. (for mols in SDF)
    """
    original: str
    errorMsg: str
    properties: dict


@dataclass
class PyMoleculeLoaderResult:
    """
    The results of a loading operation.

    Attributes:
        compoundByStr: may be blank if irrelevant. The map of source data string to the compound loaded.
        compoundList: the compounds successfully loaded.
        errorList: an error record is added here for each one we failed to load in Sapio.
    """
    compoundByStr: dict[str, PyCompound] | None
    compoundList: list[PyCompound] | None
    errorList: list[ChemLoadingError] | None


@dataclass
class ChemCompleteImportPojo:
    # Variables declaration with type hints
    dataTypeName: str
    successPartRecordIdList: list[int] | None
    successSampleRecordIdList: list[int] | None
    errors: list[ChemLoadingError] | None
    registeredOriginalParts: list[int]
    allRegisteredPartsNoDuplicates: list[int]
    newPartList: list[int]
    numOldParts: int


@dataclass
class ChemInteractiveRegisterRequestPojo:
    dataType: str
    fileType: ChemFileType
    fileDataEncodedBase64: str | None
    addingItems: bool

    def __init__(self, data_type: str, file_type: ChemFileType, is_adding_items: bool, file_data: bytes):
        self.dataType = data_type
        self.fileType = file_type
        self.addingItems = is_adding_items
        self.fileDataEncodedBase64 = base64.b64encode(file_data).decode()


@dataclass
class CompoundLoadRequestPojo:
    """
    Describes a load request for the Load Compound Endpoint.

    Attributes:
        dataType: The data type of records to be registered in Sapio.
        loadType: The source data's type you are loading.
        dataList: If the source data is not a file, here you specify a list of string describing molecule for that src.
        fileDataBase64: If the source data is a file, the file's base64 data content.
    """
    dataType: str
    loadType: ChemLoadType
    dataList: list[str] | None
    fileDataBase64: str | None

    def __init__(self, data_type: str, load_type: ChemLoadType, data_list: list[str] | None = None,
                 file_data: bytes | None = None):
        self.dataType = data_type
        self.loadType = load_type
        if load_type is ChemLoadType.SDF_FILE:
            self.dataList = None
            if file_data is None:
                raise ValueError("The file data must be specify the the load type is of a file type.")
            self.fileDataBase64 = base64.b64encode(file_data).decode()
        else:
            self.dataList = data_list
            self.fileDataBase64 = None


@dataclass
class ChemRegisterRequestPojo:
    """
    Data payload to send to webservice to request registration in RegisterCompoundEndpoint

    Attributes:
        dataType: The data type of records to be registered in Sapio.
        registrationList: This list must be of correct data structure suitable for the type. For example, for CompoundPart data type the canonical form must be resolved by earlier call.
    """
    dataType: str
    registrationList: list[PyCompound]

    def __init__(self, data_type: str, registration_list: list[PyCompound]):
        self.dataType = data_type
        self.registrationList = registration_list


@dataclass
class ImageDataRequestPojo:
    """
    Payload to request in endpoint loading of image data of a compound or a reaction in Sapio's unified drawing format.

    Attributes:
        dataList: The list of data about the images of the molecules or reactions. This can be SMILES, MOL, or INCHI. SMILES is expected. INCHI can be ambiguous to chrality as well as tautomers.
        isReaction: true if the underlying data is of RxN format of reaction.
    """
    dataList: list[str]
    reaction: bool

    def __init__(self, data_list: list[str], is_reaction: bool):
        self.dataList = data_list
        self.reaction = is_reaction


@dataclass
class PyIndigoReactionPojo:
    """
    The result of loading a reaction.

    Attributes:
        products: products of the reaction.
        reactants: reactants of the reaction.
        reactionSmiles: the SMILES representation of this reaction.
        reactionRxn: the RxN of an arotmized reaction
        reactionRenderRxn: the RxN of no-automapping, DE-aromotized reaction, with 2D coordinates of atoms computed.
    """
    products: list[PyCompound]
    reactants: list[PyCompound]
    reactionSmiles: str
    reactionRxn: str
    reactionRenderRxn: str


@dataclass
class ChemQuickSearchContextData:
    """
    Do not directly make use of this class. The only use for this class is to pass in the "next page"
    produced by the endpoint.

    When obtaining the first page, this parameter argument should not be passed at all (not created with default values).
    """
    previousPageSearchAfterJsonStack: list[str] | None
    nextPageSearchAfter: str | None
    pitId: str | None
    query: str | None
    joinSapioPartType: str | None
    simUpperLimit: float | None


@dataclass
class ChemSearchRequestPojo:
    """
    Payload to send to endpoint to request a chemical search in Sapio.
    This can be a substructure search or similarity search.

    Attributes:
        searchStr: The search string of SMILES or SMARTS you are searching
        searchType: The type of search you are doing.
        joinMethod: The registry you are using to join with Sapio record. This is not relevant for reactions.
        contextData: The context data of the current page passed to you by result of previous page. If this is the first page you are querying, leave this as None.
        simSearchUpperLimit: similarity search upper limit, between 0.0 to 1.0, valid only to similarity searches.
    """
    searchStr: str
    searchType: ChemSearchType
    joinSapioType: str | None
    contextData: ChemQuickSearchContextData | None
    simSearchUpperLimit: float | None

    def __init__(self, search_str: str, search_type: ChemSearchType, join_sapio_type: str | None = None,
                 context_data: ChemQuickSearchContextData | None = None, sim_search_upper: float | None = None):
        self.searchStr = search_str
        self.searchType = search_type
        self.joinSapioType = join_sapio_type
        self.contextData = context_data
        self.simSearchUpperLimit = sim_search_upper


@dataclass
class ChemSearchResponsePojo:
    """
    A response object of a chemistry quick search.
    NOTE: It's possible to have next page while the current records list is blank (although without additional query conditions, this should be rare).

    Attributes:
         recordsOfPage: The records returned in this page.
         nextPageAvailable: Whether there is still a next page to query. If this is filled, then the context to query next page is on nextPageContext.

    """
    recordIdListOfPage: list[int]
    nextPageAvailable: bool
    nextPageContext: ChemQuickSearchContextData


class MAFFTStrategy(Enum):
    """
    Select one of the strategies for MAFFT multi sequence alignment.
    """
    # Accuracy-Orientated options
    L_INS_i = 0
    G_INS_i = 1
    E_INS_i = 2
    # Speed-Orientated options
    AUTO = 3
    FFT_NS_i = 4
    FFT_NS_2 = 5
    FFT_NS_1 = 6
    NW_NS_i = 7
    NW_NS_2 = 8
    NW_NS_PartTree_1 = 9


class MultiSequenceAlignmentTool(Enum):
    MAFFT = 0


class MultiSeqAlignemntSeqType(Enum):
    nucleic = 0
    amino = 1


@dataclass
class MultiSequenceAlignmentSeqPojo:
    seqId: str
    seqString: str | None
    hyphSeqString: str | None

    @staticmethod
    def create(id: str, seq: str):
        ret = MultiSequenceAlignmentSeqPojo(seqId=id, seqString=seq, hyphSeqString=None)
        return ret


class MAFFTRunOptions:
    strategy: MAFFTStrategy
    noScore: bool
    op: float
    lop: float
    lep: float
    lexp: float
    LOP: float
    LEXP: float
    fmodel: bool

    def __init__(self, strategy: MAFFTStrategy = MAFFTStrategy.AUTO, noScore: bool = True,
                 op: float = 1.53, lop: float = -2.0, lep: float = 0.1, lexp: float = -0.1,
                 LOP: float = -6.0, LEXP: float = 0.00, fmodel: bool = False):
        self.strategy = strategy
        self.noScore = noScore
        self.op = op
        self.lop = lop
        self.lep = lep
        self.lexp = lexp
        self.LOP = LOP
        self.LEXP = LEXP
        self.fmodel = fmodel

    def to_json(self) -> dict[str, Any]:
        return {
            "strategy": self.strategy.name,
            "noScore": self.noScore,
            "op": self.op,
            "lop": self.lop,
            "lep": self.lep,
            "lexp": self.lexp,
            "LOP": self.LOP,
            "LEXP": self.LEXP,
            "fmodel": self.fmodel
        }


class ClustalOOptions:
    dealign: bool
    numCombinedIterations: int | None
    numGuideTreeIterations: int | None
    numHMMIterations: int | None

    def __init__(self, dealign: bool = False, numCombinedIterations: int | None = None,
                 numGuideTreeIterations: int | None = None, numHMMIterations: int | None = None):
        self.dealign = dealign
        self.numCombinedIterations = numCombinedIterations
        self.numGuideTreeIterations = numGuideTreeIterations
        self.numHMMIterations = numHMMIterations

    def to_json(self) -> dict[str, Any]:
        return {
            "dealign": self.dealign,
            "numCombinedIterations": self.numCombinedIterations,
            "numGuideTreeIterations": self.numGuideTreeIterations,
            "numHMMIterations": self.numHMMIterations
        }


class KAlignOptions:
    gapOpenPenalty: float | None
    gapExtensionPenalty: float | None
    terminalGapPenalty: float | None

    def __init__(self, gapOpenPenalty: float | None = None, gapExtensionPenalty: float | None = None,
                 terminalGapPenalty: float | None = None):
        self.gapOpenPenalty = gapOpenPenalty
        self.gapExtensionPenalty = gapExtensionPenalty
        self.terminalGapPenalty = terminalGapPenalty

    def to_json(self) -> dict[str, Any]:
        return {
            "gapOpenPenalty": self.gapOpenPenalty,
            "gapExtensionPenalty": self.gapExtensionPenalty,
            "terminalGapPenalty": self.terminalGapPenalty
        }


class MuscleOptions:
    consiters: int
    refineiters: int

    def __init__(self, consiters: int = 2, refineiters: int = 100):
        self.consiters = consiters
        self.refineiters = refineiters

    def to_json(self) -> dict[str, Any]:
        return {
            "consiters": self.consiters,
            "refineiters": self.refineiters
        }


@dataclass
class MultiSequenceAlignmentRequestPojo:
    tool: MultiSequenceAlignmentTool
    seqType: MultiSeqAlignemntSeqType
    inputSeqs: list[MultiSequenceAlignmentSeqPojo]
    parameters: dict[str, Any] | None

    def __init__(self, tool: MultiSequenceAlignmentTool, seq_type: MultiSeqAlignemntSeqType,
                 input_sequences: list[MultiSequenceAlignmentSeqPojo],
                 parameters: dict[str, Any]):
        self.tool = tool
        self.seqType = seq_type
        self.inputSeqs = input_sequences
        self.parameters = parameters


class BioFileType(Enum):
    """
    Different bio registry supported file types.
    """
    FASTA = 0
    GENBANK = 1
    PDB = 2
    CIF = 3


@dataclass
class BioFileRegistrationRequest:
    """
    A request object for a single bio-registration request on parts.
    """
    dataTypeName: str
    fileType: BioFileType
    prefilledFieldMapList: list[dict[str, Any]] | None
    overwriteExisting: bool
    fileContent: str

    def __init__(self, data_type_name: str, file_type: BioFileType, file_content: str,
                 prefilled_field_map_list: list[dict[str, Any]] | None = None, overwrite: bool = False):
        self.dataTypeName = data_type_name
        self.fileType = file_type
        self.fileContent = file_content
        self.prefilledFieldMapList = prefilled_field_map_list
        self.overwriteExisting = overwrite


@dataclass
class BioFileRegistrationResponse:
    """
    A response object for a single bio-registration request on parts.
    """
    newRecordIdList: list[int]
    oldRecordIdList: list[int]


@dataclass
class ChemExportSDFRequest:
    """
    A request to export SDF data from Sapio to Python REST client.
    """
    partDataTypeName: str
    partRecordIdList: list[int]

    forceV3000: bool
    fieldNameList: list[str] | None
    assayNameList: list[str] | None
    additionalPropertiesByRecordId: dict[int, dict[str, Any]] | None

    def __init__(self, partDataTypeName: str, partRecordIdList: list[int], forceV3000: bool = True,
                 fieldNameList: list[str] | None = None, assayNameList: list[str] | None = None,
                 additionalPropertiesByRecordId: dict[int, dict[str, Any]] | None = None):
        self.partDataTypeName = partDataTypeName
        self.partRecordIdList = partRecordIdList
        self.forceV3000 = forceV3000
        self.fieldNameList = fieldNameList
        self.assayNameList = assayNameList
        self.additionalPropertiesByRecordId = additionalPropertiesByRecordId
